import json
import time
import uuid
import httpx
import jwt
import aipkgs_notifications.singleton as singleton
import aipkgs_notifications.utils as utils
import aipkgs_notifications.response as ai_response
import aipkgs_notifications.payload as ai_payload
import aipkgs_notifications.enums as enums


APNS_HOST_URL_SANDBOX = "https://api.sandbox.push.apple.com"
APNS_HOST_URL_PROD = "https://api.push.apple.com"


class Config:
    def __init__(self, verbose: bool = None):
        self.verbose = verbose or False


@singleton.Singleton
class APNS:
    def __init__(self, key_id: str = '', team_id: str = '', bundle_id: str = '', is_prod: bool = None, p8_key_path: str = '', pem_file_path: str = '', apns_priority: int = None,
                 apns_expiration: int = None):
        self.config = Config()

        self.ALGORITHM = 'ES256'
        self.KEY_ID = key_id
        self.TEAM_ID = team_id
        self.BUNDLE_ID = bundle_id
        self.IS_PROD = is_prod
        self.AUTH_P8_KEY = p8_key_path
        self.AUTH_PEM_KEY = pem_file_path
        self.AUTH_TOKEN = None
        self.APNS_PRIORITY = str(apns_priority) or '10'
        self.APNS_EXPIRATION = str(apns_expiration) or '0'

        if self.IS_PROD:
            self.APNS_HOST_URL = APNS_HOST_URL_PROD
        else:
            self.APNS_HOST_URL = APNS_HOST_URL_SANDBOX

        # assert self.KEY_ID, "KEY_ID is null or empty"
        # assert self.TEAM_ID, "TEAM_ID is null or empty"
        # assert self.BUNDLE_ID, "BUNDLE_ID is null or empty"
        # assert self.AUTH_P8_KEY or self.AUTH_PEM_KEY, "AUTH_P8_KEY or AUTH_PEM_KEY is null or empty"

    @property
    def authentication_method(self) -> enums.AuthenticationMethod:
        if self.AUTH_P8_KEY:
            return enums.AuthenticationMethod.P8
        elif self.AUTH_PEM_KEY:
            return enums.AuthenticationMethod.PEM
        else:
            return None

    def __initialize_apns(self, key_id='', team_id='', bundle_id='', is_prod: bool = None, p8_key_path='', pem_file_path='', apns_priority: int = None,
                          apns_expiration: int = None):
        self.__init__(key_id=key_id, p8_key_path=p8_key_path, pem_file_path=pem_file_path, team_id=team_id, bundle_id=bundle_id, is_prod=is_prod, apns_priority=apns_priority,
                      apns_expiration=apns_expiration)

    def initialize_apns(self, key_id='', team_id='', bundle_id='', is_prod: bool = None, p8_key_path='', pem_file_path='', apns_priority: int = None, apns_expiration: int = None):
        self.__initialize_apns(key_id=key_id, team_id=team_id, bundle_id=bundle_id, is_prod=is_prod, p8_key_path=p8_key_path, pem_file_path=pem_file_path,
                               apns_priority=apns_priority,
                               apns_expiration=apns_expiration)

    def __generate_auth_token(self, expires: int = None) -> str:
        if self.AUTH_TOKEN:
            if expires:
                token_expired = True
                now = time.time()
                decoded_token = jwt.decode(self.AUTH_TOKEN, verify=False)
                if (decoded_token['exp'] - now) > 60:  # check if 60 seconds is left, consider it expired
                    token_expired = True
                elif decoded_token['exp'] > time.time():
                    token_expired = False
                if decoded_token['exp'] < time.time():
                    token_expired = True

                if token_expired is False:
                    return self.AUTH_TOKEN

        with open(self.AUTH_P8_KEY, 'r') as file:
            private_key = file.read()

        headers = {
            'kid': self.KEY_ID,
        }

        now = time.time()
        token_payload = {
            'iss': self.TEAM_ID,
            'iat': int(time.time()),
        }
        if expires:
            expires = now + (60 * expires)  # minutes
            token_payload['exp'] = int(expires)

        token = jwt.encode(
            token_payload,
            private_key,
            algorithm=self.ALGORITHM,
            headers=headers
        )

        self.AUTH_TOKEN = token

        return token

    def push(self, device_token: str, payload: ai_payload.Payload, push_type: enums.PushType = None, collapse_id: str = None) -> ai_response.APNSResponse:
        if self.authentication_method is None:
            raise Exception("Authentication method is not defined")

        FULL_URL = f"{self.APNS_HOST_URL}/3/device/{device_token}"

        auth_token = None
        if self.authentication_method == enums.AuthenticationMethod.P8:
            auth_token = self.__generate_auth_token(expires=None)

        bundle_id = self.BUNDLE_ID
        if push_type == enums.PushType.push_to_talk:
            bundle_id = f"{bundle_id}.voip-ptt"

        # headers
        headers = {
            "apns-id": str(uuid.uuid4()),
            "apns-push-type": push_type.value if push_type else enums.PushType.alert.value,
            "apns-expiration": self.APNS_EXPIRATION,
            "apns-priority": self.APNS_PRIORITY,
            "apns-topic": bundle_id,
            "apns-collapse-id": collapse_id,
            "apns-unix-time": str(int(time.time())),
        }
        if auth_token:
            headers["authorization"] = f"bearer {auth_token}"

        headers = utils.remove_nulls(headers)

        # print
        if self.config.verbose:
            print('---------------- sending push notification ----------------')
            print(f"headers: {json.dumps(headers, indent=4)}")
            print(f"data: {json.dumps(payload.to_dict(), indent=4)}")

        # send request
        response = None
        if self.authentication_method == enums.AuthenticationMethod.P8:
            client = httpx.Client(http2=True)
            response = client.post(
                FULL_URL,
                headers=headers,
                json=payload.to_dict(),
            )
        elif self.authentication_method == enums.AuthenticationMethod.PEM:
            client = httpx.Client(http2=True, cert=self.AUTH_PEM_KEY)
            response = client.post(
                FULL_URL,
                headers=headers,
                json=payload.to_dict(),
            )

        apns_response = ai_response.APNSResponse(httpx_response=response)

        if self.config.verbose:
            print(f"is_sent: {apns_response.is_sent()}")
            print(f"status_code: {apns_response.status_code}")
            print(f"apns_id: {apns_response.apns_id}")
            print(f"apns_unique_id:  {apns_response.apns_unique_id}")
            print(f"timestamp: {apns_response.timestamp.time if apns_response.timestamp else None}")

        return apns_response


def config() -> Config:
    return APNS.shared.config


def initialize_apns(key_id='', team_id='', bundle_id='', is_prod: bool = None, p8_key_path='', pem_file_path='', apns_priority: int = None, apns_expiration: int = None):
    APNS.shared.initialize_apns(key_id=key_id, team_id=team_id, bundle_id=bundle_id, is_prod=is_prod, p8_key_path=p8_key_path, pem_file_path=pem_file_path,
                                apns_priority=apns_priority, apns_expiration=apns_expiration)
    return APNS.shared


def push_raw(device_token: str, payload: ai_payload.Payload, push_type: enums.PushType = None, collapse_id: str = None) -> ai_response.APNSResponse:
    return APNS.shared.push(device_token=device_token, payload=payload, push_type=push_type, collapse_id=collapse_id)


def push(device_token: str, title: str, body: str = None, data: dict = None, badge: int = None, push_type: enums.PushType = None, collapse_id: str = None) -> ai_response.APNSResponse:
    alert_payload = ai_payload.AlertPayload(title=title, body=body)
    payload = ai_payload.Payload(alert=alert_payload, badge=badge, data=data)

    return push_raw(device_token=device_token, payload=payload, collapse_id=collapse_id)



import enum


class PushType(enum.Enum):
    alert = "alert"
    background = "background"
    voip = "voip"
    complication = "complication"
    file_provider = "fileprovider"
    mdm = "mdm"
    unknown = "unknown"
    push_to_talk = "pushtotalk"


class AuthenticationMethod(enum.Enum):
    P8 = "p8"  # you get the AuthKey.p8 file from the Apple Developer Portal
    PEM = "pem"  # convert the exported .p12 file to .pem file using 'openssl pkcs12 -clcerts -legacy -nodes -in Certificates.p12 -out AuthKey.pem'


class InterruptionLevel(enum.Enum):
    active = "active"
    time_sensitive = "time-sensitive"
    critical = "critical"

    def __str__(self):
        if self == InterruptionLevel.active:
            return "Active"
        elif self == InterruptionLevel.time_sensitive:
            return "Time Sensitive"
        elif self == InterruptionLevel.critical:
            return "Critical"
        else:
            return "n/a"



import aipkgs_notifications.enums as enums
import aipkgs_notifications.utils as utils


class AlertPayload:
    def __init__(self, title: str, subtitle: str = None, body: str = None):
        self.localized = False

        self.title_loc_key = None
        self.title_loc_args = None
        self.subtitle_loc_key = None
        self.subtitle_loc_args = None
        self.body_loc_key = None
        self.body_loc_args = None

        self.title = title
        self.subtitle = subtitle
        self.body = body

    def to_dict(self) -> dict:
        data = {"title": self.title,
                "subtitle": self.subtitle,
                "body": self.body}

        if self.localized:
            data["title-loc-key"] = self.title_loc_key
            data["title-loc-args"] = self.title_loc_args
            data["subtitle-loc-key"] = self.subtitle_loc_key
            data["subtitle-loc-args"] = self.subtitle_loc_args
            data["body-loc-key"] = self.body_loc_key
            data["body-loc-args"] = self.body_loc_args

        data = utils.remove_nulls(data)
        return data

    def localize_payload(self, title_loc_key: str = None, title_loc_args: list = None, subtitle_loc_key: str = None, subtitle_loc_args: list = None, body_loc_key: str = None,
                         body_loc_args: list = None) -> None:
        self.localized = True

        self.title_loc_key = title_loc_key
        self.title_loc_args = title_loc_args
        self.subtitle_loc_key = subtitle_loc_key
        self.subtitle_loc_args = subtitle_loc_args
        self.body_loc_key = body_loc_key
        self.body_loc_args = body_loc_args


class SoundPayload:
    def __init__(self, name: str, critical: bool = None, volume: int = None):
        self.name = name
        self.critical = critical
        self.volume = volume

    def to_dict(self) -> dict:
        data = {"name": self.name,
                "critical": 1 if self.critical else None,
                "volume": self.volume if (self.volume and ((self.volume >= 0) and (self.volume <= 1))) else None}

        data = utils.remove_nulls(data)

        return data


class Payload:
    def __init__(self, alert: AlertPayload = None, category: str = None, sound: SoundPayload = None, badge: int = None, data: dict = None, thread_id: str = None,
                 content_available: bool = None, mutable_content: bool = None, target_content_id: str = None,
                 target_content_type: str = None, target_content_url: str = None, interruption_level: enums.InterruptionLevel = None, relevance_score: int = None,
                 filter_criteria: str = None,
                 stale_date: str = None, content_state: dict = None, timestamp: int = None, event: str = None, dismissal_date: int = None):
        self.alert = alert
        self.category = category
        self.sound = sound
        self.badge = badge
        self.data = data
        self.thread_id = thread_id
        self.content_available = content_available
        self.mutable_content = mutable_content
        self.target_content_id = target_content_id
        self.target_content_type = target_content_type
        self.target_content_url = target_content_url
        self.interruption_level = interruption_level
        self.relevance_score = relevance_score
        self.filter_criteria = filter_criteria
        self.stale_date = stale_date
        self.content_state = content_state
        self.timestamp = timestamp
        self.event = event
        self.dismissal_date = dismissal_date

    def to_dict(self) -> dict:
        data = {"aps": {"alert": self.alert.to_dict() if self.alert else None,
                        "category": self.category,
                        "sound": self.sound.to_dict() if self.sound else {"name": "default"},
                        "badge": self.badge,
                        "thread-id": self.thread_id,
                        "content-available": 1 if self.content_available else None,
                        "mutable-content": 1 if self.mutable_content else None,
                        "target-content-id": self.target_content_id,
                        "target-content-type": self.target_content_type,
                        "target-content-url": self.target_content_url,
                        "interruption-level": self.interruption_level.value if self.interruption_level else None,
                        "relevance-score": self.relevance_score if (self.relevance_score and ((self.relevance_score >= 0) and (self.relevance_score <= 1))) else None,
                        "filter-criteria": self.filter_criteria,
                        "stale-date": self.stale_date,
                        "content-state": self.content_state,
                        "timestamp": self.timestamp,
                        "event": self.event,
                        "dismissal-date": self.dismissal_date
                        },
                "data": self.data}

        # if self.push_type == apns.PushType.background:
        #     data["aps"]["content-available"] = 1

        if self.content_available:
            data["aps"]["alert"] = None
            data["aps"]["sound"] = None
            data["aps"]["badge"] = None

        data = utils.remove_nulls(data)

        return data



import datetime
import enum

import httpx


class APNSResponse:
    def __init__(self, httpx_response: httpx.Response):
        self.httpx_response = httpx_response

        headers = self.httpx_response.headers
        request_headers = self.httpx_response.request.headers

        timestamp = request_headers.get('apns-unix-time', None)

        self.status_code = self.httpx_response.status_code
        self.apns_id = headers.get('apns-id', None)
        self.apns_unique_id = headers.get('apns-unique-id', None)
        self.timestamp = Time(int(timestamp) * 1000) if timestamp else None

    @property
    def is_sent(self) -> bool:
        return self.httpx_response.is_success


class Time:
    def __init__(self, timestamp):
        self.timestamp = timestamp

    @property
    def time(self):
        return datetime.datetime.fromtimestamp(self.timestamp / 1000)


class Reason(enum.Enum):
    bad_collapse_id = "BadCollapseId"
    bad_device_token = "BadDeviceToken"
    bad_expiration_date = "BadExpirationDate"
    bad_message_id = "BadMessageId"
    bad_priority = "BadPriority"
    bad_topic = "BadTopic"
    device_token_not_for_topic = "DeviceTokenNotForTopic"
    duplicate_headers = "DuplicateHeaders"
    idle_timeout = "IdleTimeout"
    invalid_push_type = "InvalidPushType"
    missing_device_token = "MissingDeviceToken"
    missing_topic = "MissingTopic"
    payload_empty = "PayloadEmpty"
    topic_disallowed = "TopicDisallowed"
    bad_certificate = "BadCertificate"
    bad_certificate_environment = "BadCertificateEnvironment"
    expired_provider_token = "ExpiredProviderToken"
    forbidden = "Forbidden"
    invalid_provider_token = "InvalidProviderToken"
    missing_provider_token = "MissingProviderToken"
    bad_path = "BadPath"
    method_not_allowed = "MethodNotAllowed"
    unregistered = "Unregistered"
    payload_too_large = "PayloadTooLarge"
    too_many_provider_token_updates = "TooManyProviderTokenUpdates"
    too_many_requests = "TooManyRequests"
    internal_server_error = "InternalServerError"
    service_unavailable = "ServiceUnavailable"
    shutdown = "Shutdown"
